// session
// cs161 assignment 2 (fall 2005)
//
// kiran
//
// implementation (see header)
//

#include "logind.h"
#include "imap.h"
#include "session.h"
#include "file_io.h"


extern logind server;



// constructor
//
session::session(io_event_loop& ioloop, iohandle ioh, term_handler *t)
  : net_handler<>(ioloop, ioh),
    term(*t)
{
    term << endl << "connection from " << get_socket().getpeername() << endl;

    set_owned(true);
    file_io = dying = authenticated = false;
    get_socket().set_blocking(false);

    // if we're maxed on fds, refuse this connection
    server.cur_fds++;
    if (server.cur_fds > server.max_fds)
    {
        // TODO: this write() doesn't go through; need to do this?
        string resp = "IMAP server has too many clients. Try again later.\r\n";
        write(resp);
        term << "refusing connection" << endl;
        delete this;
    }
    else
    {
        // send greeting
        string resp = "OK logind ready.";
        send_untagged(resp);
    }
}



// destructor
//
session::~session()
{
    term << "closing connection to " << get_socket().getpeername() << endl;
    server.cur_fds--;
}



// fio_response()
//
// called by an fio_handler when the file i/o operation that was
// requested completes and the response arrives
// incoming_data()
//
void session::fio_response(fio_rsp *rsp)
{
    string s, *p;


    // sanity check
    if (!file_io)
    {
        term << "unexpected file i/o response received" << endl;
        delete rsp;
        return;
    }


    // check what kind of request this response is for
    switch (rsp->req_type)
    {
      case fio_req::LOGIN:
        if (rsp->ok)
        {
            authenticated = true;
            send_ok(rsp->tag, IMAP::LOGIN);
        }
        else
        {
            user_name = "";
            send_no(rsp->tag, IMAP::LOGIN);
        }
        break;

      case fio_req::LIST:
        for (int i = 0;  i < rsp->resp.size();  i++)
        {
            p = rsp->resp.at(i);
            s = "LIST ";
            s += *p;
            delete p;
            send_untagged(s);
        }
        if (rsp->ok)
            send_ok(rsp->tag, IMAP::LIST);
        else
            send_bad(rsp->tag, IMAP::LIST);
        break;

      default:
        term << "unknown file i/o response received" << endl;
        break;
    }

    file_io = false;
    delete rsp;


    // if a LOGOUT came while we were waiting, die now
    if (dying)
    {
        send_ok(cur_tag, IMAP::LOGOUT);
        delete this;
    }
}



// incoming_data()
//
// invoked whenever bytes are newly available on the stream; expects
// IMAP commands as LF terminated strings. processes the commands
//
int session::incoming_data(constbuf cbuf)
{
    const char  *cs;
    string      buf, line;
    int         length = 0, consumed = 0;


    // if we're dying, do nothing
    if (dying)
        return 0;


    // extract a line from the input buffer
    buf = cbuf;
    while (get_token(buf, '\n', line, consumed))
    {
        // there must be a tag
        if (!get_token(line, ' ', cur_tag, length))
        {
            send_bad(line);
            continue;
        }

        cs = line.c_str();


        // CAPABILITY
        //
        if (strncasecmp(cs, (IMAP::CAPABILITY).c_str(), 10) == 0)
        {
            string resp = "CAPABILITY IMAP4rev1";
            send_untagged(resp);
            send_ok(cur_tag, IMAP::CAPABILITY);
        }


        // LIST
        //
        else if (strncasecmp(cs, (IMAP::LIST).c_str(), 4) == 0)
        {
            string      ref, mb;

            if (!authenticated)
                send_bad(cur_tag, "not logged in.");

            else
            {
                // discard the command itself
                get_token(line, ' ', ref, length);

                // get the reference name and mailbox name
                if (!get_token(line, ' ', ref, length))
                    send_bad(cur_tag, "parameters.");
                else
                {
                    mb = line;

                    // send the file i/o helper the request
                    if (!list_req(user_name, ref, mb))
                        send_no(cur_tag, IMAP::LIST, "internal error.");
                }
            }
        }


        // LOGIN
        //
        else if (strncasecmp(cs, (IMAP::LOGIN).c_str(), 5) == 0)
        {
            string      pass;

            if (authenticated)
                send_ok(cur_tag, IMAP::LOGIN, "already logged in.");

            else
            {
                // discard the command itself
                get_token(line, ' ', pass, length);

                // get the user name and password
                if (!get_token(line, ' ', user_name, length))
                    send_bad(cur_tag, "parameters.");
                else
                {
                    pass = line;

                    // send the file i/o helper the request
                    if (!login_req(user_name, pass))
                        send_no(cur_tag, IMAP::LOGIN, "internal error.");
                }
            }
        }


        // LOGOUT
        //
        else if (strncasecmp(cs, (IMAP::LOGOUT).c_str(), 6) == 0)
        {
            string resp = "BYE logging out";
            send_untagged(resp);

            // if we're waiting for file i/o, die later
            if (file_io)
                dying = true;
            else
            {
                send_ok(cur_tag, IMAP::LOGOUT);
                delete this;
            }
        }


        // NOOP
        //
        else if (strncasecmp(cs, (IMAP::NOOP).c_str(), 4) == 0)
        {
            // TODOlater: check for anything to send untagged

            send_ok(cur_tag, IMAP::NOOP);
        }


        // unknown
        //
        else
            send_bad(cur_tag);
    }


    // TODOlater: handle case where we get nothing usable; count frequency
    // and timeout?
    // inactivity timer?


    return consumed;
}	



// end_data
//
// invoked when the peer has closed the connection; cleans up the
// session and dies
//
void session::end_data(constbuf buf)
{
    // if we're waiting for file i/o, wait till it arrives to die
    if (file_io)
    {
        dying = true;
        return;
    }

    delete this;
}



// send_ok()
//
// respond with OK to a received command
//
void session::send_ok(string& tag, string& command, string text)
{
    string resp = tag + " OK " + command + " " + text + "\r\n";
    write(resp);
}



// send_untagged()
//
// send an untagged response
//
void session::send_untagged(string& s)
{
    string resp = "* " + s + "\r\n";
    write(resp);
}



// send_bad()
//
// respond with BAD to a received command
//
void session::send_bad(string& recvd, string text)
{
    string resp = recvd + " BAD " + text + "\r\n";
    write(resp);
}



// send_no()
//
// respond with NO to a received command
//
void session::send_no(string& tag, string& cmd, string text)
{
    string resp = tag + " NO " + cmd + " " + text + "\r\n";
    write(resp);
}



// get_token()
//
// returns true if separator was found in buf; places all data upto
// the separator in token, increments consumed by the number of bytes
// upto and including the separator, and updates buf to point after
// the separator
//
bool session::get_token(string& buf, char separator,
                string& token, int& consumed)
{
    int incr = 1;

    // find the separator
    int pos = buf.find_first_of(separator);
    if (pos == string::npos)
        return false;

    // special case for removing '\r' when looking for '\n'
    if (separator == '\n'  &&  buf[pos - 1] == '\r')
    {
        pos--;
        incr = 2;
    }

    // extract the token
    token = string(buf, 0, pos);

    // skip over the separator
    pos += incr;

    // remove it from the passed buffer
    consumed += pos;
    buf = string(buf, pos, buf.size() - pos);

    return true;
}



// login_req()
//
// prepares and sends a file i/o request for LOGIN verification
//
bool session::login_req(string& user, string& pass)
{
    fio_req *req = prep_fio_req(fio_req::LOGIN, user);
    req->password = pass;
    return send_fio_req(req);
}



// list_req()
//
// prepares and sends a file i/o request for LIST
//
bool session::list_req(string& user, string& ref, string& mb)
{
    fio_req *req = prep_fio_req(fio_req::LIST, user);
    req->largs.refname = ref;
    req->largs.mbname = mb;
    return send_fio_req(req);
}



// prep_fio_req()
//
// allocates and initializes a file i/o request
//
fio_req *session::prep_fio_req(int req_type, string& user)
{
    fio_req *req = new fio_req;

    req->type = req_type;
    req->username = user;
    req->requester = this;
    req->tag = cur_tag;

    return req;
}



// send_fio_req()
//
// sends a file i/o request
//
bool session::send_fio_req(fio_req *req)
{
    fio_handler *fioh = server.get_fioh();
    if (fioh->request(req))
        return (file_io = true);

    return false;
}

